# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.

import logging

import pytest

from fillmore.scrubber import (
    build_scrub_cookies,
    build_scrub_query_string,
    _get_target_dicts,
    scrub,
    Scrubber,
    Rule,
)


@pytest.mark.parametrize(
    "value, expected",
    [
        # String values are scrubbed--this is straight-forward
        ("abc", "[Scrubbed]"),
        # Non-string values are also scrubbed--this can be problematic and engineers
        # setting up scrubbing should be thoughtful about what they're scrubbing
        (0, "[Scrubbed]"),
        ({"foo": "bar"}, "[Scrubbed]"),
    ],
)
def test_scrub(value, expected):
    assert scrub(value) == expected


@pytest.mark.parametrize(
    "cookies, keys, expected",
    [
        # Empty cookies variations: empty string, empty dict, empty list
        ("", [], ""),
        ({}, [], {}),
        ([], [], []),
        # Cover the variations of cookies as a string
        (
            "foo=bar; foo2=bar2",
            ["code", "state"],
            "foo=bar; foo2=bar2",
        ),
        (
            "code=abc123; state=base64string",
            ["code", "state"],
            "code=[Scrubbed]; state=[Scrubbed]",
        ),
        # If the cookies has nothing that needs scrubbing in it, then it returns the
        # cookies unaltered
        (
            "foo=bar;    foo2=bar2",
            ["code", "state"],
            "foo=bar;    foo2=bar2",
        ),
        # If the cookies has things that need scrubbing, then it will normalize
        # the cookies values.
        #
        # FIXME(willkg): This changes the cookies and will make debugging difficult, but
        # I don't see a better way to scrub without inadvertently fixing invalid things.
        (
            "code=abc123;    randomthing=value",
            ["code", "state"],
            "code=[Scrubbed]; randomthing=value",
        ),
        # Cover cookies as a dict
        (
            {"foo": "bar"},
            ["code", "state"],
            {"foo": "bar"},
        ),
        (
            {"code": "asdf", "foo": "bar"},
            ["code", "state"],
            {"code": "[Scrubbed]", "foo": "bar"},
        ),
        # Cover cookies as a list of tuples
        (
            [("foo", "bar")],
            ["code", "state"],
            [("foo", "bar")],
        ),
        (
            [("code", "asdf"), ("foo", "bar")],
            ["code", "state"],
            [("code", "[Scrubbed]"), ("foo", "bar")],
        ),
    ],
)
def test_scrub_cookies(cookies, keys, expected):
    scrub_fun = build_scrub_cookies(params=keys)
    assert scrub_fun(cookies) == expected


@pytest.mark.parametrize(
    "qs, keys, expected",
    [
        # Empty query_string variations: empty string, empty dict, empty list
        ("", [], ""),
        ({}, [], {}),
        ([], [], []),
        # Cover the variations of query_string as a string
        (
            "foo=bar&foo2=bar2",
            ["code", "state"],
            "foo=bar&foo2=bar2",
        ),
        (
            "code=abc123&state=base64String",
            ["code", "state"],
            "code=%5BScrubbed%5D&state=%5BScrubbed%5D",
        ),
        (
            "code=abc123&state",
            ["code", "state"],
            "code=%5BScrubbed%5D&state=%5BScrubbed%5D",
        ),
        # If the query_string has nothing that needs scrubbing in it, then
        # it returns the query_string unaltered
        (
            "test=%A&random&invalid_utf8=%A0%A1",
            ["code", "state"],
            "test=%A&random&invalid_utf8=%A0%A1",
        ),
        # If the query_string has things that need scrubbing, then it will normalize
        # the query_string values.
        #
        # FIXME(willkg): This changes the query_string and will make debugging difficult,
        # but I don't see a better way to scrub without inadvertently fixing invalid
        # things.
        (
            "code=55&test=%A&random&invalid_utf8=%A0%A1",
            ["code", "state"],
            "code=%5BScrubbed%5D&test=%25A&random=&invalid_utf8=%EF%BF%BD%EF%BF%BD",
        ),
        # Cover query_string as a dict
        (
            {"foo": "bar"},
            ["code", "state"],
            {"foo": "bar"},
        ),
        (
            {"code": "asdf", "foo": "bar"},
            ["code", "state"],
            {"code": "[Scrubbed]", "foo": "bar"},
        ),
        (
            {"foo": "bar", "state": ""},
            ["code", "state"],
            {"foo": "bar", "state": "[Scrubbed]"},
        ),
        # Cover query_string as a list of tuples
        (
            [("foo", "bar")],
            ["code", "state"],
            [("foo", "bar")],
        ),
        (
            [("code", "asdf"), ("foo", "bar")],
            ["code", "state"],
            [("code", "[Scrubbed]"), ("foo", "bar")],
        ),
        (
            [("foo", "bar"), ("state", "")],
            ["code", "state"],
            [("foo", "bar"), ("state", "[Scrubbed]")],
        ),
    ],
)
def test_scrub_query_string(qs, keys, expected):
    scrub_fun = build_scrub_query_string(params=keys)
    assert scrub_fun(qs) == expected


@pytest.mark.parametrize(
    "event, path, expected",
    [
        # Test empty things
        (
            {},
            [],
            [{}],
        ),
        # Test key traversal
        (
            {"foo": "bar"},
            [],
            [{"foo": "bar"}],
        ),
        (
            {"foo1": {"foo2": "bar"}},
            ["foo1"],
            [{"foo2": "bar"}],
        ),
        # Test [] array traversal
        (
            {
                "stack": {
                    "frames": [
                        {"index": 0, "values": {"code_id": "abcd"}},
                        {"index": 1, "values": {"state": "state_abcd"}},
                        {"index": 2, "values": {"code_id": "2_abcd"}},
                    ]
                }
            },
            ["stack", "frames", "[]", "values"],
            [
                {"code_id": "abcd"},
                {"state": "state_abcd"},
                {"code_id": "2_abcd"},
            ],
        ),
    ],
)
def test_get_target_paths(event, path, expected):
    assert list(_get_target_dicts(event, path)) == expected


@pytest.mark.parametrize(
    "event, path, expected",
    [
        # Test case where path points to item not in structure
        (
            {"exception": {"values": [{"foo": "bar"}]}},
            ["exception", "values", "[]", "stacktrace", "frames", "[]", "values"],
            [],
        )
    ],
)
def test_get_target_paths_missing(event, path, expected):
    assert list(_get_target_dicts(event, path)) == expected


class TestScrubber:
    @pytest.mark.parametrize(
        "rules, event, expected",
        [
            ([], {}, {}),
            (
                [Rule(path="foo", keys=["bar"], scrub=scrub)],
                {"foo": {"bar": "somevalue"}, "foo2": "othervalue"},
                {"foo": {"bar": "[Scrubbed]"}, "foo2": "othervalue"},
            ),
            (
                [Rule(path="foo", keys=["bar"], scrub="scrub")],
                {"foo": {"bar": "somevalue"}, "foo2": "othervalue"},
                {"foo": {"bar": "[Scrubbed]"}, "foo2": "othervalue"},
            ),
            (
                [
                    Rule(
                        path="frames.[].vars",
                        keys=["code_id", "state"],
                        scrub=scrub,
                    ),
                ],
                {
                    "frames": [
                        {"vars": {"foo": "bar"}},
                        {"vars": {"index": 4, "code_id": "abcd", "state": "def"}},
                    ],
                    "function": "somefunc",
                },
                {
                    "frames": [
                        {"vars": {"foo": "bar"}},
                        {
                            "vars": {
                                "index": 4,
                                "code_id": "[Scrubbed]",
                                "state": "[Scrubbed]",
                            }
                        },
                    ],
                    "function": "somefunc",
                },
            ),
            # Handle case where the parent isn't a dict so the key doesn't exist. This let's
            # us support a possible structure variation of request.data which could be a
            # data structure or a string.
            (
                [Rule(path="request.data", keys=["bar"], scrub=scrub)],
                {"request": {"data": "abcde"}},
                {"request": {"data": "abcde"}},
            ),
            (
                [Rule(path="request.data", keys=["bar"], scrub=scrub)],
                {"request": {"data": {"bar": "abcde"}}},
                {"request": {"data": {"bar": "[Scrubbed]"}}},
            ),
        ],
    )
    def test_scrubbing(self, rules, event, expected):
        scrubber = Scrubber(rules=rules)
        assert scrubber(event, {}) == expected

    def test_scrub_error(self, caplog):
        """Test scrub error when no error_handler is specified"""

        def bad_scrub(value):
            raise Exception("scruberror")

        event = {"request": {"data": {"foo": "bar"}}}

        scrubber = Scrubber(
            rules=[
                Rule(
                    path="request",
                    keys=["data"],
                    scrub=bad_scrub,
                )
            ]
        )
        scrubber(event, {})

        # The event should have ERROR WHEN SCRUBBING instead of [Scrubbed] or
        # something else
        assert event == {"request": {"data": "ERROR WHEN SCRUBBING"}}

        # A message should be logged at the ERROR level
        assert caplog.record_tuples == [
            (
                "fillmore.scrubber",
                logging.ERROR,
                "scrub fun error: bad_scrub, error: scruberror",
            )
        ]

    def test_scrub_error_error_handler(self, caplog):
        """Test scrub error when error_handler is specified"""

        class ErrorHandler:
            def __init__(self):
                self.errors = []

            def __call__(self, msg):
                self.errors.append(msg)

        handler = ErrorHandler()

        def bad_scrub(value):
            raise Exception("scruberror")

        event = {"request": {"data": {"foo": "bar"}}}

        scrubber = Scrubber(
            rules=[
                Rule(
                    path="request",
                    keys=["data"],
                    scrub=bad_scrub,
                )
            ],
            error_handler=handler,
        )
        scrubber(event, {})

        # The event should have ERROR WHEN SCRUBBING instead of [Scrubbed] or
        # something else
        assert event == {"request": {"data": "ERROR WHEN SCRUBBING"}}

        # error_handler was called with message
        assert handler.errors == ["scrub fun error: bad_scrub, error: scruberror"]

        # A message should be logged at the ERROR level
        assert caplog.record_tuples == [
            (
                "fillmore.scrubber",
                logging.ERROR,
                "scrub fun error: bad_scrub, error: scruberror",
            )
        ]

    def test_scrub_error_error_handler_error(self, caplog):
        """Test scrub error when error_handler is specified and error_handler kicks up error"""

        def bad_error_handler(msg):
            raise Exception("handlererror")

        def bad_scrub(value):
            raise Exception("scruberror")

        event = {"request": {"data": {"foo": "bar"}}}

        scrubber = Scrubber(
            rules=[
                Rule(
                    path="request",
                    keys=["data"],
                    scrub=bad_scrub,
                )
            ],
            error_handler=bad_error_handler,
        )
        scrubber(event, {})

        # The event should have ERROR WHEN SCRUBBING instead of [Scrubbed] or
        # something else
        assert event == {"request": {"data": "ERROR WHEN SCRUBBING"}}

        # Multiple messages should be logged at the ERROR level
        assert caplog.record_tuples == [
            (
                "fillmore.scrubber",
                logging.ERROR,
                "scrub fun error: bad_scrub, error: scruberror",
            ),
            (
                "fillmore.scrubber",
                logging.ERROR,
                "error in error_handler bad_error_handler",
            ),
        ]

    def test_path_error(self, caplog):
        """Test path error when error_handler is not specified"""

        event = {"request": {"data": {"foo": "bar"}}}

        scrubber = Scrubber(
            rules=[
                Rule(
                    path="request.[].data",
                    keys=["foo"],
                    scrub="scrub",
                )
            ],
        )
        scrubber(event, {})

        # The event isn't altered by rules with bad paths
        assert event == {"request": {"data": {"foo": "bar"}}}

        # Multiple messages should be logged at the ERROR level
        assert caplog.record_tuples == [
            (
                "fillmore.scrubber",
                logging.ERROR,
                "scrubber error: error: path 'request.[]' doesn't match event structure",
            ),
        ]

    def test_path_error_error_handler(self, caplog):
        """Test path error when error_handler is specified"""

        class ErrorHandler:
            def __init__(self):
                self.errors = []

            def __call__(self, msg):
                self.errors.append(msg)

        handler = ErrorHandler()

        event = {"request": {"data": {"foo": "bar"}}}

        scrubber = Scrubber(
            rules=[
                Rule(
                    path="request.[].data",
                    keys=["foo"],
                    scrub="scrub",
                )
            ],
            error_handler=handler,
        )
        scrubber(event, {})

        # The event isn't altered by rules with bad paths
        assert event == {"request": {"data": {"foo": "bar"}}}

        # error_handler was called with message
        assert handler.errors == [
            "scrubber error: error: path 'request.[]' doesn't match event structure"
        ]

        # Multiple messages should be logged at the ERROR level
        assert caplog.record_tuples == [
            (
                "fillmore.scrubber",
                logging.ERROR,
                "scrubber error: error: path 'request.[]' doesn't match event structure",
            ),
        ]

    def test_path_error_error_handler_error(self, caplog):
        """Test path error when error_handler is specified and errors"""

        def bad_error_handler(msg):
            raise Exception("handlererror")

        event = {"request": {"data": {"foo": "bar"}}}

        scrubber = Scrubber(
            rules=[
                Rule(
                    path="request.[].data",
                    keys=["foo"],
                    scrub="scrub",
                )
            ],
            error_handler=bad_error_handler,
        )
        scrubber(event, {})

        # The event isn't altered by rules with bad paths
        assert event == {"request": {"data": {"foo": "bar"}}}

        # Multiple messages should be logged at the ERROR level
        assert caplog.record_tuples == [
            (
                "fillmore.scrubber",
                logging.ERROR,
                "scrubber error: error: path 'request.[]' doesn't match event structure",
            ),
            ("fillmore.scrubber", 40, "error in error_handler bad_error_handler"),
        ]



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.

import pytest

from fillmore.test import SentryTestHelper


@pytest.fixture
def sentry_helper() -> SentryTestHelper:
    """SentryTestHelper fixture

    This creates a helper and sets up the context so it's usable immediately.

    """
    return SentryTestHelper()



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.


"""Fillmore is a Python library for scrubbing Sentry events."""

# x.y.z or x.y.z.dev0
__version__ = "0.1.1"


SCRUBBER_MODULE_NAME = __name__



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.

# Configuration file for the Sphinx documentation builder.
#
# This file only contains a selection of the most common options. For a full
# list see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

# -- Path setup --------------------------------------------------------------

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.
#
import os
import sys

sys.path.insert(0, os.path.abspath("."))

cwd = os.getcwd()
project_root = os.path.dirname(cwd)
src_root = os.path.join(project_root, "src")
sys.path.insert(0, src_root)


# -- Project information -----------------------------------------------------

project = "Fillmore"
copyright = "2022, Will Kahn-Greene"
author = "Will Kahn-Greene"


# -- General configuration ---------------------------------------------------

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom
# ones.
extensions = [
    "sphinx.ext.autodoc",
]

# Add any paths that contain templates here, relative to this directory.
templates_path = ["_templates"]

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This pattern also affects html_static_path and html_extra_path.
exclude_patterns = ["_build", "Thumbs.db", ".DS_Store"]

pygments_style = "sphinx"

autodoc_typehints = "description"
autoclass_content = "both"
autodoc_default_options = {
    "class-doc-from": "both",
    "member-order": "bysource",
    "inheireted-members": True,
}


# -- Options for HTML output -------------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
#
html_theme = "sphinx_rtd_theme"

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
html_static_path = ["_static"]



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.

import pytest
import sentry_sdk
from sentry_sdk.integrations.stdlib import StdlibIntegration

from fillmore.scrubber import Scrubber, Rule
from fillmore.test import SentryTestHelper, get_sentry_base_url


@pytest.mark.parametrize(
    "url, expected",
    [
        ("http://localhost/0", "http://localhost/"),
        ("http://localhost:8000/0", "http://localhost:8000/"),
        ("http://foo:bar@localhost:8000/0", "http://localhost:8000/"),
    ],
)
def test_get_sentry_base_url(url, expected):
    assert get_sentry_base_url(url) == expected


def test_helper_capture_events():
    """Test that helper captures events."""
    helper = SentryTestHelper()
    with helper.init() as sentry_client:
        assert sentry_client.events == []
        try:
            raise Exception("intentional")
        except Exception as exc:
            sentry_sdk.capture_exception(exc)

        (event,) = sentry_client.events
        assert event["exception"]["values"][0]["type"] == "Exception"
        assert event["exception"]["values"][0]["value"] == "intentional"

        try:
            raise Exception("another intentional")
        except Exception as exc:
            sentry_sdk.capture_exception(exc)

        (event1, event2) = sentry_client.events
        assert event1["exception"]["values"][0]["type"] == "Exception"
        assert event1["exception"]["values"][0]["value"] == "intentional"

        assert event2["exception"]["values"][0]["type"] == "Exception"
        assert event2["exception"]["values"][0]["value"] == "another intentional"


def test_helper_capture_exceptions_without_stack():
    scrubber = Scrubber(
        rules=[
            Rule(
                path="exception.values.[].stacktrace.frames.[].vars",
                keys=["username"],
                scrub="scrub",
            )
        ]
    )

    helper = SentryTestHelper()
    with helper.init(before_send=scrubber) as sentry_client:
        assert sentry_client.events == []

        sentry_sdk.capture_exception(Exception("intentional"))

        (event,) = sentry_client.events
        assert event["exception"]["values"][0]["type"] == "Exception"
        assert event["exception"]["values"][0]["value"] == "intentional"

        sentry_sdk.capture_exception(Exception("another intentional"))

        (event1, event2) = sentry_client.events
        assert event1["exception"]["values"][0]["type"] == "Exception"
        assert event1["exception"]["values"][0]["value"] == "intentional"

        assert event2["exception"]["values"][0]["type"] == "Exception"
        assert event2["exception"]["values"][0]["value"] == "another intentional"


def test_helper_contexts():
    """Test that new context clears events."""
    helper = SentryTestHelper()
    with helper.init() as sentry_client:
        assert sentry_client.events == []
        try:
            raise Exception("intentional")
        except Exception as exc:
            sentry_sdk.capture_exception(exc)

        (event,) = sentry_client.events
        assert event["exception"]["values"][0]["type"] == "Exception"
        assert event["exception"]["values"][0]["value"] == "intentional"

    with helper.init() as sentry_client:
        assert sentry_client.events == []


def test_helper_reset():
    """Test reset clears events."""
    helper = SentryTestHelper()
    with helper.init() as sentry_client:
        assert sentry_client.events == []
        try:
            raise Exception("intentional")
        except Exception as exc:
            sentry_sdk.capture_exception(exc)

        (event,) = sentry_client.events
        assert event["exception"]["values"][0]["type"] == "Exception"
        assert event["exception"]["values"][0]["value"] == "intentional"

        sentry_client.reset()
        assert sentry_client.events == []


def test_helper_reuse():
    helper = SentryTestHelper()

    with sentry_sdk.Hub(None):
        # Initialize sentry
        sentry_sdk.init(
            dsn="http://user:pwd@localhost:8000/0",
            auto_enabling_integrations=False,
            default_integrations=False,
            integrations=[StdlibIntegration()],
        )

        # Initialize a sentry with no integrations and verify it has no integrations
        kwargs = {
            "auto_enabling_integrations": False,
            "default_integrations": False,
            "integrations": [],
        }
        with helper.init(**kwargs) as sentry_client:
            assert sentry_client.events == []
            try:
                raise Exception("intentional")
            except Exception as exc:
                sentry_sdk.capture_exception(exc)

            (event,) = sentry_client.events
            assert event["sdk"]["integrations"] == []

        # Try reusing the sentry client we created already and verify it has integrations
        with helper.reuse() as reused_client:
            assert reused_client.events == []
            try:
                raise Exception("intentional")
            except Exception as exc:
                sentry_sdk.capture_exception(exc)

            (event,) = reused_client.events
            assert event["sdk"]["integrations"] == ["stdlib"]


def test_capture_events_with_scrubber():
    """Test the helper with scrubber."""
    scrubber = Scrubber(
        rules=[
            Rule(
                path="exception.values.[].stacktrace.frames.[].vars",
                keys=["username"],
                scrub="scrub",
            )
        ]
    )

    helper = SentryTestHelper()
    with helper.init(before_send=scrubber) as sentry_client:
        assert sentry_client.events == []
        try:
            username = "foo"  # noqa
            raise Exception("intentional")
        except Exception as exc:
            sentry_sdk.capture_exception(exc)

        (event,) = sentry_client.events
        error = event["exception"]["values"][0]
        assert error["type"] == "Exception"
        assert error["value"] == "intentional"
        # Verify the username frame-local var is scrubbed
        assert error["stacktrace"]["frames"][0]["vars"]["username"] == "[Scrubbed]"



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.

"""Utility functions for setting up Sentry."""

import logging
from typing import Any, Callable, List

import sentry_sdk
from sentry_sdk.integrations.logging import ignore_logger

from fillmore import SCRUBBER_MODULE_NAME


logger = logging.getLogger(__name__)


def set_up_sentry(
    sentry_dsn: str,
    release: str,
    host_id: str,
    integrations: List[Any] = None,
    before_send: Callable = None,
    **kwargs: Any,
) -> None:
    """Set up Sentry

    By default, this will set up default integrations
    (https://docs.sentry.io/platforms/python/configuration/integrations/default-integrations/),
    but not the auto-enabling ones.

    :param sentry_dsn: the Sentry DSN
    :param release: the release name to tag events with
    :param host_id: some str representing the host this service is running on
    :param integrations: list of sentry integrations to set up;
    :param before_send: set this to a callable to handle the Sentry before_send hook

        For scrubbing, do something like this::

            scrubber = Scrubber(rules=SCRUB_RULES_DEFAULT + my_scrub_rules)

        and then pass that as the ``before_send`` value.

    :param kwargs: any additional arguments to pass to sentry_sdk.init()

    """
    if not sentry_dsn:
        return

    sentry_sdk.init(
        dsn=sentry_dsn,
        release=release,
        send_default_pii=False,
        server_name=host_id,
        # This prevents Sentry from trying to enable all the auto-enabling
        # integrations. We only want the ones we explicitly set up. This
        # prevents sentry from loading the Falcon integration (which fails) in a Django
        # context.
        auto_enabling_integrations=False,
        integrations=integrations or [],
        before_send=before_send or None,
        **kwargs,
    )

    # Ignore logging from this module
    ignore_logger(SCRUBBER_MODULE_NAME)



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.

import importlib
import logging
from urllib.parse import parse_qsl, urlencode
from typing import Any, Callable, Generator, List, Optional, Union

import attrs


LOGGER = logging.getLogger(__name__)


MASK_TEXT: str = "[Scrubbed]"


ALL_COOKIE_KEYS: Any = object()
ALL_QUERY_STRING_KEYS: Any = object()


def scrub(value: str) -> str:
    """Scrub a value"""
    return MASK_TEXT


def build_scrub_cookies(params: List[str]) -> Callable:
    """Scrub specified keys in HTTP request cookies

    Sentry says the cookies can be:

    * an unparsed string
    * a dictionary
    * a list of tuples

    For the unparsed string, this parses it and figures things out.

    For dictionary and list of tuples, this returns the scrubbed forms of those.

    If the specified params is ``ALL_COOKIE_KEYS``, then this will scrub all
    cookie values.

    """

    def _scrub_cookies(value: Union[str, dict, list]) -> Union[str, dict, list]:
        to_scrub = params

        if not value:
            return value

        if isinstance(value, dict):
            if to_scrub is ALL_COOKIE_KEYS:
                value = {key: MASK_TEXT for key in value.keys()}
                return value

            for param in to_scrub:
                if param in value:
                    value[param] = MASK_TEXT
            return value

        if isinstance(value, list):
            if to_scrub is ALL_COOKIE_KEYS:
                value = [(pair[0], MASK_TEXT) for pair in value]
                return value

            for i, pair in enumerate(value):
                if pair[0] in to_scrub:
                    value[i] = (pair[0], MASK_TEXT)
            return value

        has_scrubbed_item = False
        scrubbed_pairs = []
        for cookie in value.split(";"):
            name, val = cookie.split("=", 1)
            name = name.strip()
            val = val.strip()

            if to_scrub is ALL_COOKIE_KEYS or name in to_scrub:
                if val:
                    val = MASK_TEXT
                    has_scrubbed_item = True
            scrubbed_pairs.append((name, val))

        if not has_scrubbed_item:
            return value

        return "; ".join(["=".join(pair) for pair in scrubbed_pairs])

    return _scrub_cookies


def build_scrub_query_string(params: List[str]) -> Callable:
    """Scrub specified keys in an HTTP request query_string

    Sentry says the query_string can be:

    * an unparsed string
    * a dictionary
    * a list of tuples

    For the unparsed string, this parses it and figures things out. If there's nothing
    that needs to be scrubbed, then it returns the original string. Otherwise it
    returns a query_string value with the items scrubbed, and reformed into a
    query_string. This sometimes means that other things in the string have changed and
    that may make debugging issues a little harder.

    For dictionary and list of tuples, this returns the scrubbed forms of those.

    If the params is ``ALL_QUERY_STRING_KEYS``, then this will scrub all
    query_string values.

    .. Note::

       The Sentry docs say that the query_string could be part of the url. This doesn't
       handle that situation.

    """

    def _scrub_query_string(value: Union[str, list, dict]) -> Union[str, list, dict]:
        to_scrub = params
        if not value:
            return value

        if isinstance(value, dict):
            if to_scrub is ALL_QUERY_STRING_KEYS:
                value = {key: MASK_TEXT for key in value.keys()}
                return value

            for param in to_scrub:
                if param in value:
                    value[param] = MASK_TEXT
            return value

        if isinstance(value, list):
            if to_scrub is ALL_QUERY_STRING_KEYS:
                value = [(pair[0], MASK_TEXT) for pair in value]
                return value

            for i, pair in enumerate(value):
                if pair[0] in to_scrub:
                    value[i] = (pair[0], MASK_TEXT)
            return value

        has_scrubbed_item = False
        scrubbed_pairs = []
        for name, val in parse_qsl(value, keep_blank_values=True):
            if to_scrub is ALL_QUERY_STRING_KEYS or name in to_scrub:
                val = MASK_TEXT
                has_scrubbed_item = True
            scrubbed_pairs.append((name, val))

        if not has_scrubbed_item:
            return value

        return urlencode(scrubbed_pairs)

    return _scrub_query_string


def str2list(s: str) -> List[str]:
    """Splits a string into parts using "." as the separator"""
    return s.split(".")


class RuleError(Exception):
    pass


def thing2fun(thing: Union[Callable, str]) -> Callable:
    """Convert str or Callable to a Callable

    If the thing refers to a scrub function in this module, return that.

    If the thing is a dotted Python path to some other function, return that.

    :raise RuleError: if the thing is not a callable or does not exist

    """
    if callable(thing):
        return thing

    if isinstance(thing, str):
        if thing in globals():
            # If it's global in this module, then pull that
            fn = globals()[thing]
            if callable(fn):
                return fn

        elif "." in thing:
            module_name, class_name = thing.rsplit(".", 1)
            module = importlib.import_module(module_name)
            try:
                fn = getattr(module, class_name)
            except AttributeError:
                raise RuleError(f"{thing} does not exist")

            if callable(fn):
                return fn

    raise RuleError(f"{thing} is not a callable or a string or does not exist")


@attrs.define
class Rule:
    """

    :param path: Python dotted path of key names with ``[]`` to denote
        arrays to traverse pointing to a dict with values to scrub.

    :param keys: list of keys to scrub values of

    :param scrub: is a callable that takes a value and returns a scrubbed value.
        For example:

        .. code-block::

           def hide_letter_a(value>: str) -> str:
               return "".join([letter if letter != "a" else "*" for letter in value])


    Rule example::

        Rule(
            path="request.data",
            keys=["csrfmiddlewaretoken"],
            scrub=scrub,
        )

        Rule(
            path="request.data",
            keys=["csrfmiddlewaretoken"],
            scrub="somemodule.scrubfunction",
        )

    """

    path: List[str] = attrs.field(converter=str2list)
    keys: List[str]
    scrub: Callable = attrs.field(converter=thing2fun)


SCRUB_RULES_DEFAULT: List[Rule] = [
    # Hide "username" and "password" in stacktrace frame-local vars
    Rule(
        path="exception.values.[].stacktrace.frames.[].vars",
        keys=["username", "password"],
        scrub=scrub,
    ),
]


class RulePathError(Exception):
    """The rule path doesn't match the structure of the event"""


def _get_target_dicts(event: dict, path: List[str]) -> Generator[dict, None, None]:
    """Given a path, yields the target dicts.

    Keys should be dict keys. To traverse all the items in an array value, use ``[]``.

    With this event::

        {
            "request": { ... },
            "exception": {
                "stacktrace": {
                    "frames": [
                        {"name": "frame1", "vars": { ... }},
                        {"name": "frame2", "vars": { ... }},
                        {"name": "frame3", "vars": { ... }},
                        {"name": "frame4", "vars": { ... }},
                    ]
                }
            }
        }

    Example path values::

        ["request"]
        ["exception", "stacktrace", "frames", "[]", "vars"]

    """
    parent = event
    for i, part in enumerate(path):
        if part == "[]":
            if not isinstance(parent, (tuple, list)):
                # FIXME(willkg): raising an issue here means that this rule is
                # misconfigured but we don't end up scrubbing anything, so it
                # could result in leaked data and that seems bad
                partial_path = ".".join(path[0 : i + 1])
                raise RulePathError(
                    f"path {partial_path!r} doesn't match event structure"
                )

            for item in parent:
                yield from _get_target_dicts(item, path[i + 1 :])
            return

        elif part in parent:
            parent = parent[part]

        else:
            # This path doesn't point to a thing in the structure
            return

    if isinstance(parent, dict):
        yield parent


class Scrubber:
    """Scrubber pipeline for Sentry events

    This is used as a ``before_send`` value as discussed here:

    https://docs.sentry.io/platforms/python/configuration/filtering/

    You create a :py:class:`fillmore.scrubber.Scrubber` with a list of scrub
    rules. When sentry_sdk is about to emit an event, the
    :py:class:`fillmore.scrubber.Scrubber` applies the scrub rules to the event
    and returns the scrubbed event data.

    If a scrub rule kicks up an error, then the configured ``error_handler`` is
    called.

    """

    def __init__(
        self,
        rules: List[Rule] = SCRUB_RULES_DEFAULT,
        error_handler: Optional[Callable] = None,
    ):
        """
        :param rules: list of Rule instances
        :param error_handler: function that takes a msg (str) and is called
            when either a scrub rule or getting the specified key by path in the
            Sentry event kicks up an error; this lets you emit some kind of signal
            when the Sentry scrubbing code is failing so it doesn't do it silently.

            By default, this logs an exception.

        """
        self.rules = rules
        self.error_handler = error_handler

    def __call__(self, event: dict, hint: Any) -> dict:
        """Implements before_send function interface and scrubs Sentry event

        This tries really hard to be very defensive such that even if there are bugs in
        the scrubs, it still emits something to Sentry.

        It will log errors, so we should look for those log statements. They'll
        all be coming from the "fillmore.scrubber" logger.

        """

        for rule in self.rules:
            try:
                for parent in _get_target_dicts(event, rule.path):
                    if not parent:
                        continue

                    for key in rule.keys:
                        if key not in parent:
                            continue

                        val = parent[key]

                        try:
                            filtered_val = rule.scrub(val)
                        except Exception as inner_exc:
                            msg = f"scrub fun error: {rule.scrub.__name__}, error: {inner_exc}"
                            LOGGER.exception(msg)
                            if self.error_handler is not None:
                                try:
                                    self.error_handler(msg)
                                except Exception:
                                    LOGGER.exception(
                                        f"error in error_handler {self.error_handler.__name__}"
                                    )

                            filtered_val = "ERROR WHEN SCRUBBING"

                        parent[key] = filtered_val

            except Exception as outer_exc:
                msg = f"scrubber error: error: {outer_exc}"
                LOGGER.exception(msg)
                if self.error_handler is not None:
                    try:
                        self.error_handler(msg)
                    except Exception:
                        LOGGER.exception(
                            f"error in error_handler {self.error_handler.__name__}"
                        )

        return event



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.

from contextlib import contextmanager
import json
import logging
from pathlib import Path
from typing import Any, Dict, Generator, List
from unittest.mock import patch
from urllib.parse import urlparse
import uuid

import sentry_sdk
from sentry_sdk.transport import Transport

from fillmore.scrubber import Scrubber


LOGGER = logging.getLogger(__name__)


def get_sentry_base_url(sentry_dsn: str) -> str:
    """Given a sentry_dsn, returns the base url

    This is helpful for tests that need the url to the fakesentry api.

    :param sentry_dsn: the sentry base url

    :raises TypeError: when sentry_dsn is not a string

    :raises ValueError: when sentry_dsn is empty

    """
    if not isinstance(sentry_dsn, str):
        raise TypeError("sentry_dsn must be a str")

    if not sentry_dsn:
        raise ValueError("sentry_dsn is required")

    parsed_dsn = urlparse(sentry_dsn)
    netloc = parsed_dsn.netloc
    if "@" in netloc:
        netloc = netloc[netloc.find("@") + 1 :]

    return f"{parsed_dsn.scheme}://{netloc}/"


class _CaptureTransport(Transport):
    """Sentry transport that captures emitted events."""

    def __init__(self) -> None:
        Transport.__init__(self)
        self._queue = None

        self.events: List[Dict[Any, Any]] = []

    def capture_event(self, event: Dict[Any, Any]) -> None:
        self.events.append(event)

    def capture_envelope(self, envelope: Any) -> None:
        pass

    def reset(self) -> None:
        self.events = []


class ReuseException(Exception):
    """Raised when there's no sentry_sdk client configured to reuse"""


class SentryTestHelper:
    """Sentry test helper for initializing Sentry and capturing events.

    This helper lets you create new sentry_sdk clients or reuse existing
    configured ones.

    You can access emitted events with the ``.events`` attribute.

    You can reset the event list with ``.reset()``.

    """

    def __init__(self) -> None:
        self._transport = _CaptureTransport()

    @property
    def events(self) -> List[Dict[Any, Any]]:
        """Access the event list."""
        return self._transport.events

    def reset(self) -> None:
        """Resets the event list."""
        self._transport.reset()

    @contextmanager
    def init(
        self, *args: Any, **kwargs: Any
    ) -> Generator["SentryTestHelper", None, None]:
        """Create a new sentry_sdk client with specified args

        This creates a new sentry_sdk client with the specified args and
        patches the client transport with one that captures Sentry events that
        are being emitted. This lets you assert things against events.

        Arguments are the same as to sentry_sdk.Client.

        .. seealso::

           https://docs.sentry.io/platforms/python/configuration/options/

        """
        with sentry_sdk.Hub(None):
            hub = sentry_sdk.Hub.current
            client = sentry_sdk.Client(*args, **kwargs)
            hub.bind_client(client)

            self._transport.reset()
            client.transport = self._transport
            yield self

    @contextmanager
    def reuse(self) -> Generator["SentryTestHelper", None, None]:
        """Re-use the current sentry_sdk client, but patch the transport

        This clears the breadcrumbs of the current sentry_sdk scope and patches
        the client transport with one that captures Sentry events that are
        being emitted. This lets you assert things against events.

        :raises ReuseException: if there's no sentry client initialized

        """
        client = sentry_sdk.Hub.current.client
        if not client:
            raise ReuseException("there is no client to reuse")

        self._transport.reset()

        # Clear the breadcrumbs in the scope
        sentry_sdk.Hub.current.scope.clear_breadcrumbs()

        # Mock the transport with one that captures events
        with patch.object(client, attribute="transport", new=self._transport):
            yield self


class ConfigurationError(Exception):
    pass


class SaveEvents:
    """Utility wrapper for saving Sentry events to files on disk.

    This is for collecting Sentry event data to build tests to verify scrubbing
    is working as you need it to.

    .. Note::

       Capturing Sentry event data and writing tests against that is fragile
       and not as good as writing integration tests that kick up Sentry events
       that then get scrubbed.

       Make sure to update captured data periodically. This will avoid skew
       from sentry_sdk updates where they change the shape of the events or
       what's included in events as well as changes to your code which changes
       frame-local vars, context data, and so on.

    Usage::

        scrubber = Scrubber( ... )
        scrubber = SaveEvents(
            wrapped_scrubber=scrubber,
            outputdir="/some/path"
        )

    """

    def __init__(self, wrapped_scrubber: Scrubber, outputdir: str):
        self.wrapped_scrubber = wrapped_scrubber
        self.outputdir = Path(outputdir)
        if not self.outputdir.is_dir():
            raise ConfigurationError(f"outputdir {outputdir} does not exist")

    def __call__(self, event: dict, hint: Any) -> dict:
        try:
            event_id = uuid.uuid4().hex
            path = self.outputdir / f"{event_id}.json"
            data = json.dumps(event)
            path.write_text(data)
        except Exception as exc:
            LOGGER.exception(f"error in SaveEvents.__call__: {exc}")

        return self.wrapped_scrubber(event=event, hint=hint)



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.

"""
Test Scrubber with DjangoIntegration.
"""

import pytest
from sentry_sdk.integrations.django import DjangoIntegration
from werkzeug.test import Client

from fillmore.scrubber import (
    build_scrub_cookies,
    build_scrub_query_string,
    Scrubber,
    Rule,
)
from tests.django.myapp.wsgi import application


INIT_KWARGS = {
    "auto_enabling_integrations": False,
    "integrations": [DjangoIntegration()],
}


@pytest.fixture
def client():
    return Client(application)


def test_scaffolding(sentry_helper, client):
    """Verifies the scaffolding for django integration tests is working"""
    with sentry_helper.init(**INIT_KWARGS) as sentry_client:
        resp = client.get("/broken")
        assert resp.status_code == 500

        # Enforces there's only one event and unpacks it
        (event,) = sentry_client.events

        assert "django" in event["sdk"]["integrations"]
        assert "request" in event


def test_scrub_request_headers(sentry_helper, client):
    """Test scrubbing headers"""
    scrubber = Scrubber(
        rules=[
            Rule(path="request.headers", keys=["Auth-Token"], scrub="scrub"),
        ],
    )
    kwargs = dict(INIT_KWARGS)
    kwargs["before_send"] = scrubber
    with sentry_helper.init(**kwargs) as sentry_client:
        resp = client.get("/broken", headers=[("Auth-Token", "abcde")])
        assert resp.status_code == 500

        # Enforces there's only one event and unpacks it
        (event,) = sentry_client.events

        assert event["request"]["headers"]["Auth-Token"] == "[Scrubbed]"
        assert event["request"]["headers"]["Host"] == "localhost"


def test_scrub_request_querystring(sentry_helper, client):
    """Test scrubbing querystring which is in the query_string field as a single string"""
    scrubber = Scrubber(
        rules=[
            Rule(
                path="request",
                keys=["query_string"],
                scrub=build_scrub_query_string(params=["code", "state"]),
            ),
        ],
    )
    kwargs = dict(INIT_KWARGS)
    kwargs["before_send"] = scrubber
    with sentry_helper.init(**kwargs) as sentry_client:
        resp = client.get(
            "/broken", query_string={"code": "foo", "state": "bar", "color": "pink"}
        )
        assert resp.status_code == 500

        # Enforces there's only one event and unpacks it
        (event,) = sentry_client.events

        query_string = list(sorted(event["request"]["query_string"].split("&")))
        assert query_string == [
            "code=%5BScrubbed%5D",
            "color=pink",
            "state=%5BScrubbed%5D",
        ]


def test_scrub_request_cookies(sentry_helper, client):
    """Test scrubbing cookies

    NOTE(willkg): Cookie data is only sent when send_default_pii=True and it
    gets sent in both the request.headers.Cookie field as well as the
    request.cookies field.

    """
    scrubber = Scrubber(
        rules=[
            Rule(
                path="request.headers",
                keys=["Cookie"],
                scrub=build_scrub_cookies(params=["csrftoken", "sessionid"]),
            ),
            Rule(
                path="request",
                keys=["cookies"],
                scrub=build_scrub_cookies(params=["csrftoken", "sessionid"]),
            ),
        ]
    )
    kwargs = dict(INIT_KWARGS)
    kwargs["send_default_pii"] = True
    kwargs["before_send"] = scrubber
    with sentry_helper.init(**kwargs) as sentry_client:
        # Add a bunch of cookies
        client.set_cookie(server_name="localhost", key="csrftoken", value="abcde")
        client.set_cookie(server_name="localhost", key="sessionid", value="someid")
        client.set_cookie(server_name="localhost", key="foo", value="bar")

        resp = client.get("/broken")
        assert resp.status_code == 500

        # Enforces there's only one event and unpacks it
        (event,) = sentry_client.events

        cookie_header = list(sorted(event["request"]["headers"]["Cookie"].split("; ")))
        assert cookie_header == [
            "csrftoken=[Scrubbed]",
            "foo=bar",
            "sessionid=[Scrubbed]",
        ]
        assert event["request"]["cookies"] == {
            "csrftoken": "[Scrubbed]",
            "foo": "bar",
            "sessionid": "[Scrubbed]",
        }



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.

import os

from django.core.wsgi import get_wsgi_application

os.environ.setdefault("DJANGO_SETTINGS_MODULE", "tests.django.myapp.settings")

application = get_wsgi_application()



#!/usr/bin/env python

# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.


import os
import re
from setuptools import find_packages, setup


def get_version():
    fn = os.path.join("src", "fillmore", "__init__.py")
    vsre = r"""^__version__ = ['"]([^'"]*)['"]"""
    version_file = open(fn).read()
    return re.search(vsre, version_file, re.M).group(1)


def get_file(fn):
    with open(fn) as fp:
        return fp.read()


INSTALL_REQUIRES = [
    "attrs>=21.2.0",
    "sentry-sdk>=1.5.0",
]


setup(
    name="fillmore",
    version=get_version(),
    description="Sentry event scrubber and utilities library",
    long_description=(get_file("README.rst") + "\n\n" + get_file("HISTORY.rst")),
    author="Will Kahn-Greene",
    author_email="willkg@mozilla.com",
    url="https://github.com/willkg/fillmore",
    install_requires=INSTALL_REQUIRES,
    packages=find_packages(where="src"),
    package_dir={"": "src"},
    include_package_data=True,
    license="MPLv2",
    zip_safe=False,
    keywords="sentry scrubber",
    classifiers=[
        "Development Status :: 4 - Beta",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: Mozilla Public License 2.0 (MPL 2.0)",
        "Natural Language :: English",
        "Operating System :: OS Independent",
        "Programming Language :: Python :: 3 :: Only",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Topic :: Software Development :: Libraries :: Python Modules",
    ],
    project_urls={
        "Documentation": "https://fillmore.readthedocs.io/",
        "Tracker": "https://github.com/willkg/fillmore/issues",
        "Source": "https://github.com/willkg/fillmore/",
    },
    entry_points={
        "pytest11": ["fillmore=fillmore.pytest_plugin"],
    },
    options={"bdist_wheel": {"universal": "1"}},
)
